#include "file.h"

/* Should be in furrycat/file.h */
/* check that a (texture) filename is valid on Windows */
int invalid_windows_filename(unsigned char *buf, unsigned short len) {
  unsigned short i;
  int seen_colon, seen_backslash;
  
  seen_colon = seen_backslash = 0;

  for (i = 0; i < len; i++) {
    if (! isprint(buf[i])) return 1;
    /* allowed in drive assignment */
    if (buf[i] == ':') if (seen_colon++) return 1;
    if (buf[i] == '/') return 1;
    if (buf[i] == '*') return 1;
    if (buf[i] == '?') return 1;
    if (buf[i] == '"') return 1;
    if (buf[i] == '<') return 1;
    if (buf[i] == '>') return 1;
    if (buf[i] == '|') return 1;
  }

  return 0;
}

/* find the full path to a(ny) file with given basename */
int find_file(char *dirname, char *basename, char *actual) {
  if (! dirname) return 1;
  if (! basename) return 2;

  if (! dirname[0]) dirname = ".";
#ifdef _MSC_VER
  WIN32_FIND_DATA find;
  find.dwFileAttributes = FILE_ATTRIBUTE_NORMAL;
  if (! SetCurrentDirectory(dirname)) {
    int error = GetLastError();
    if (error != ERROR_PATH_NOT_FOUND) fprintf(stderr, "Can't open %s directory: %d\n", dirname, error);
    return 3;
  }
  HANDLE found = FindFirstFile(basename, &find);
  if (found == INVALID_HANDLE_VALUE) return 4;
  if (! ulMakePath(actual, dirname, find.cFileName)) {
    fprintf(stderr, "Can't construct path %s in %s\n", find.cFileName, dirname);
    FindClose(found);
    return 5;
  }
  FindClose(found);
  return 0;
#else
  ulDir *directory = ulOpenDir(dirname);
  if (! directory) {
    fprintf(stderr, "Can't read directory %s\n", dirname);
    return 4;
  }

  struct ulDirEnt *dir;
  while (dir = ulReadDir(directory)) {
    if (str_equals(dir->d_name, ".") || str_equals(dir->d_name, "..")) continue;
    if (dir->d_isdir) {
      char *path = new char[UL_NAME_MAX + 1];
      if (! ulMakePath(path, dirname, dir->d_name)) {
        fprintf(stderr, "Can't construct path %s in %s\n", dir->d_name, dirname);
        delete[] path;
        continue;
      }
      int ret = find_file(path, basename, actual);
      delete[] path;
      if (ret) continue;
      ulCloseDir(directory);
      return 0;
    }
    /* file */
    if (str_equiv(dir->d_name, basename)) {
      if (actual) {
        if (! ulMakePath(actual, dirname, dir->d_name)) {
          fprintf(stderr, "Found file %s but can't make path from %s\n", dir->d_name, dirname);
          ulCloseDir(directory);
          return 2;
        }
      }
      ulCloseDir(directory);
      return 0;
    }
    //printf("%s: %s\n", dirname, dir->d_name);
  }
  ulCloseDir(directory);
  return 3;
#endif
}

/* find dirname and put it in dest pointer*/
void set_dirname(char *filename, unsigned short filename_length, char **dirname, unsigned short *dirname_length) {
  size_t i;

  /* hmm, do I look like a Sierra employee? */
  if (! dirname) return;

  if (! filename_length) filename_length = strlen(filename);
  for (i = filename_length; i >= 0; i--) {
    if (filename[i] == '\\' || filename[i] == '/') break;
  }
  if (! i) i = filename_length;
  if (dirname_length) *dirname_length = (unsigned short) i;
  *dirname = new char[i + 1];
  memmove(*dirname, filename, i);
  (*dirname)[i] = '\0';
}

/* find basename and put it in dest pointer*/
void set_basename(char *filename, unsigned short filename_length, char **basename, unsigned short *basename_length) {
  size_t i;

  /* hmm, do I look like a Sierra employee? */
  if (! basename) return;

  if (! filename_length) filename_length = strlen(filename);
  for (i = 0; i < filename_length; i++) {
    if (filename[filename_length - i] == '\\' || filename[filename_length - i] == '/') {
      i--;
      break;
    }
  }
  if (basename_length) *basename_length = (unsigned short) i;
  *basename = filename + filename_length - i;
}
