#include "gl_texture.h"

/* Use IL! */
extern char *cwd;
extern int textures_enabled;

extern GLuint background_texture;
static GLuint background_ok = 0;
extern int show_background;

int load_background(char *filename) {
  GLuint new_background_texture;
  ILuint ret;

  /* Allocate new background */
  glGenTextures(1, &new_background_texture);

  /* Load the file */
  printf("Loading background image '%s': ", filename);
  ret = load_gl_image(&new_background_texture, filename);
  if (ret) {
    printf(" DevIL error code %u\n", ret);
    return 2;
  }
  printf("%u\n", new_background_texture);

  printf("Freeing old background: ");
  if (background_ok) {
    glDeleteTextures(1, &background_texture);
    printf("%u\n", background_texture);
  }
  else printf("no need\n");

  background_texture = new_background_texture;
  glBindTexture(GL_TEXTURE_2D, background_texture);

  background_ok = 1;

  return 0;
}

void cleanup_gl_textures(texture_data_t &gl_textures) {
  if (! gl_textures.num_textures) return;
  for (GLuint i = 0; i < gl_textures.num_textures; i++) {
    glDeleteTextures(1, &gl_textures.textures[i]);
  }
  delete[] gl_textures.textures; gl_textures.textures = 0;
  delete[] gl_textures.ok; gl_textures.ok = 0;
  gl_textures.num_textures = 0;
}

/* Load textures using DevIL */
int load_textures(unsigned short num_textures, texture_t *textures, texture_data_t &gl_textures, char *dirname) {
  unsigned int i;

  if (! textures_enabled) {
    fprintf(stderr, "Textures are disabled: not even trying to load any!\n");
    return num_textures;
  }

  gl_textures.num_textures = (GLuint) num_textures;
  gl_textures.textures = new GLuint[num_textures];
  gl_textures.ok = new GLuint[num_textures];

  int errors = 0;
  glGenTextures(num_textures, &gl_textures.textures[0]);
  for (i = 0; i < num_textures; i++) {
    texture_t &texture = textures[i];
    char *filename = new char[texture.name_length + 1];
    memmove(filename, texture.name, texture.name_length);
    filename[texture.name_length] = '\0';

    /* Get basename */
    char *basename = get_basename(filename);

    /* Try to load the file */
    if (! load_image(dirname, basename, &gl_textures.textures[i])) {
      gl_textures.ok[i] = 1;
      glBindTexture(GL_TEXTURE_2D, gl_textures.textures[i]);
      continue;
    }

    /* Try actual path */
    if (! load_gl_image(&gl_textures.textures[i], filename)) {
      printf("Loaded %s\n", filename);
      gl_textures.ok[i] = 1;
      glBindTexture(GL_TEXTURE_2D, gl_textures.textures[i]);
      continue;
    }

    /* Can't load it */
    gl_textures.ok[i] = 0;
    errors++;
    printf("Can't load texture \"%s\"\n", basename);
  }

  return errors;
}
