#include "gripper.h"
#include "grips.h"

int read_grippers(FILE *file, unsigned short *num_grippers, gripper_t **grippers) {
  if (read_uint16(file, num_grippers)) return 1;
  if (! *num_grippers) return 0;
  *grippers = new gripper_t[*num_grippers];
  for (unsigned short i = 0; i < *num_grippers; i++) {
    gripper_t &gripper = (*grippers)[i];
    if (read_uint16(file, &gripper.name_length)) return 2;
    gripper.name = new char[gripper.name_length + 1];
    if (read_arbitrary(file, gripper.name, gripper.name_length)) return 3;
    gripper.name[gripper.name_length] = '\0';
    sgMakeIdentMat4(gripper.local);
    for (unsigned short j = 0; j < 4; j++) {
      sgVec3 tmp;
      if (read_vec3(file, &tmp)) return j + 2;
      sgCopyVec3(gripper.local[j], tmp);
    }
    /*
    sgTransposeNegateMat4(gripper.local);
    gripper.local[3][0] *= -1.0;
    gripper.local[3][1] *= -1.0;
    gripper.local[3][2] *= -1.0;
    */
    /*
    sgMat4 m;
    sgMultMat4(m, gripper.local, xgl);
    sgCopyMat4(gripper.local, m);
    */
    if (read_uint16(file, &gripper.bone_id)) return 6;
  }
  return 0;
}

int write_grippers(FILE *file, unsigned short *num_grippers, gripper_t **grippers) {
  if (write_uint16(file, num_grippers)) return 1;
  for (unsigned short i = 0; i < *num_grippers; i++) {
    gripper_t &gripper = (*grippers)[i];
    if (write_uint16(file, &gripper.name_length)) return 2;
    if (write_arbitrary(file, gripper.name, gripper.name_length)) return 3;
    gripper.name[gripper.name_length] = '\0';
    for (unsigned short j = 0; j < 4; j++) {
      sgVec3 tmp;
      sgCopyVec3(tmp, gripper.local[j]);
      if (write_vec3(file, &tmp)) return j + 2;
    }
    if (write_uint16(file, &gripper.bone_id)) return 6;
  }
  return 0;
}

void cleanup_grippers(unsigned short num_grippers, gripper_t *grippers) {
  if (! num_grippers || ! grippers) return;
  for (unsigned short i = 0; i < num_grippers; i++) {
    gripper_t &gripper = grippers[i];
    if (gripper.name_length && gripper.name) delete[] gripper.name;
    gripper.name_length = 0; gripper.name = 0;
  }
  delete[] grippers;
}

/* Make default gripper */
static int make_default_gripper(gripper_t &gripper, unsigned char *gripper_coords, unsigned char *gripper_name, unsigned short gripper_bone_id) {
  /* Make transform */
  if (gripper_coords) {
    for (int i = 0; i < 4; i++) {
      memmove(gripper.local[i], gripper_coords + i * 12, 12);
    }
  }

  /* Set name */
  gripper.name_length = GRIPPER_NAME_LENGTH;
  gripper.name = new char[gripper.name_length + 1];
  memmove(gripper.name, gripper_name, gripper.name_length);
  gripper.name[gripper.name_length] = '\0';

  /* Set bone ID */
  gripper.bone_id = gripper_bone_id;
  printf("Making gripper '%s' with bone %u at (% 0.4f, % 0.4f, % 0.4f)\n", gripper.name, gripper.bone_id, gripper.local[3][0], gripper.local[3][1], gripper.local[3][2]);

  return 0;
}

/* Make default SWT grippers */
int make_swt_gripper(gripper_t &gripper, unsigned short gripper_id) {
  if (gripper_id >= SWT_NUM_GRIPPERS) return 1;
  return make_default_gripper(gripper, swt_gripper_coords[gripper_id], swt_gripper_name[gripper_id], swt_gripper_bone_id[gripper_id]);
}

/* Make default TAC grippers */
int make_tac_gripper(gripper_t &gripper, unsigned short gripper_id) {
  if (gripper_id >= TAC_NUM_GRIPPERS) return 1;
  return make_default_gripper(gripper, tac_gripper_coords[gripper_id], tac_gripper_name[gripper_id], tac_gripper_bone_id[gripper_id]);
}

/* Make default BF1 grippers */
int make_bf1_gripper(gripper_t &gripper, unsigned short gripper_id) {
  if (gripper_id >= BF1_NUM_GRIPPERS) return 1;
  return make_default_gripper(gripper, bf1_gripper_coords[gripper_id], bf1_gripper_name[gripper_id], bf1_gripper_bone_id[gripper_id]);
}

/* Make grips for a gun */
int make_gun_grip(gripper_t &grip, unsigned short gripper_id) {
  return make_default_gripper(grip, 0, gun_grip_name[gripper_id], 0);
}
