#include "version.h"
#include "gui.h"
#include "video.h"

extern int loading;
extern char *cwd;
extern std::string meshdir, animdir, texturedir, bgdir, screendir;
extern unsigned long meshindex, animindex, textureindex, bgindex, screenindex;
extern GLfloat xrot, yrot, zrot;
extern GLfloat xpos, ypos, zpos;
extern sgVec4 rotquat;
extern gsm_t *character, *weapon;
extern gsa_t *character_animation, *weapon_animation;
extern int culling;
extern int reverse_normals;
extern int show_hit_detect;
extern int show_background;
extern int textures_enabled;
extern int material_hack;
extern int wireframe;
extern int exitasap;
extern int show_mesh;
extern int show_bones;
extern int show_normals;
extern int antialiasing;
extern int frame;
extern int paused;
extern SDL_SysWMinfo wm;
#ifdef WIN32
static char gskin_filename[MAX_PATH];
static OPENFILENAME filesel;
static HINSTANCE instance;
/* our HWND is in wm.window */
static HWND menu_window = 0;
static HWND textures_window = 0;
static HWND mesh_window = 0;
static HWND animation_window = 0;
extern int fake_animation;

#ifdef WINDOWS2000_TRANSPARENCY_DOES_NOT_SUCK
/* DLL handles for transparency */
static shared_library_t library = 0;
static function_pointer_t function = 0;
#endif
#endif
static int reloading_textures = 0;
static int fileselector_in_use = 0;

#ifdef WIN32
int gui_fileselector(const char *ext, char *filter, char *title, const char *dir, unsigned long *index) {
  int ret;

  if (fileselector_in_use) {
    fprintf(stderr, "THREAD SAFETY: Fileselector is in use!\n");
    return 111;
  }

  fileselector_in_use = 1;
  gskin_filename[0] = '\0';
  memset(&filesel, 0, sizeof(filesel));

  filesel.lStructSize = sizeof(filesel);
  filesel.hwndOwner = 0;
  filesel.lpstrFilter = filter;
  filesel.lpstrFile = gskin_filename;
  filesel.lpstrInitialDir = dir;
  filesel.lpstrTitle = title;
  if (index) filesel.nFilterIndex = *index;
  filesel.nMaxFile = MAX_PATH;
  filesel.Flags = OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY;
  filesel.lpstrDefExt = 0;

  /* Save or load selector? */
  if (ext) {
    filesel.lpstrDefExt = ext;
    ret = GetSaveFileName(&filesel);
  }
  else ret = GetOpenFileName(&filesel);

  if (index) *index = filesel.nFilterIndex;

  fileselector_in_use = 0;
  return ret;
}

static void make_transparent(HWND window) {
#ifdef WINDOWS2000_TRANSPARENCY_DOES_NOT_SUCK
  /* Here we WOULD show the window with transparency but it flickers with GL */
  if (! window) return;
  if (function) {
    SetWindowLongPtr(window, GWL_EXSTYLE, GetWindowLongPtr(window, GWL_EXSTYLE) | WS_EX_LAYERED);
    SetLayeredWindowAttributes(window, 0, 200, LWA_ALPHA);
  }
#endif
}
#endif

void gui_set_dirname(unsigned short *len, char **dirname) {
#ifdef WIN32
  *len = (unsigned short) filesel.nFileOffset - 1;
  *dirname = new char[*len + 1];
  memmove(*dirname, filesel.lpstrFile, filesel.nFileOffset);
  (*dirname)[*len] = '\0';
#endif
}

void gui_load_mesh() {
#ifdef WIN32
  if (gui_fileselector(0, "Model files\0*.gsm;*.ms3d\0GSkinMesh files\0*.gsm\0Officer models\0SWT_*.gsm\0Kid models\0KID_*.gsm\0Male character models\0TAC_*.gsm;BG1_*.gsm;BG2_*.gsm;BG3_*.gsm;MH1_*.gsm;MH2_*.gsm\0Female character models\0BF1_*.gsm;FH1_*.gsm;FH2_*.gsm;PRH_*.gsm\0Milkshape 3D files\0*.ms3d\0All files\0*.*\0\0", "Load GSKinMesh", meshdir.c_str(), &meshindex)) {
    gsm_t *gsm = new gsm_t;
    gsm->dirname = new char[filesel.nFileOffset];
    memmove(gsm->dirname, filesel.lpstrFile, filesel.nFileOffset);
    gsm->dirname[filesel.nFileOffset - 1] = '\0';
    gsm->dirname_length = filesel.nFileOffset - 1;
    meshdir = gsm->dirname;
    memset(&gsm->gl_textures, 0, sizeof(gsm->gl_textures));

    int ret;

    /* Try Milkshape if it's ms3d */
    printf("Loading mesh: %s\n", filesel.lpstrFile);
    if (str_equiv(file_extension(filesel.lpstrFile), "ms3d")) {
      ms3d_model_t *ms3d = new ms3d_model_t;
      ret = read_ms3d(*ms3d, filesel.lpstrFile);
      if (ret) {
        fprintf(stderr, "Error loading Milkshape model: %d\n", ret);
        cleanup_ms3d(ms3d);
      }
      else {
        //describe_ms3d(*ms3d, MS3D_DESCRIBE_ALL);
        ret = ms3d_to_gsm(*gsm, *ms3d);
        if (ret) {
          fprintf(stderr, "Can't convert Milkshape model: %d\n", ret);
        }
        cleanup_ms3d(ms3d);
      }
    }
    else {
      ret = read_gsm(*gsm, filesel.lpstrFile);
      if (ret) {
        fprintf(stderr, "Error loading GSkinMesh: %d\n", ret);
      }
    }

    if (ret) return;

    SetDlgItemText(mesh_window, IDC_MESH_NAME, filesel.lpstrFile + filesel.nFileOffset);
    SetDlgItemInt(mesh_window, IDC_NUM_POLYGONS, gsm->num_triangles, 1);
    SetDlgItemInt(mesh_window, IDC_NUM_BONES, gsm->num_bones, 1);
    SetDlgItemText(mesh_window, IDC_CLASS, gsm->type);
    ret = load_textures(gsm->num_textures, gsm->textures, gsm->gl_textures, gsm->dirname);
    if (ret) {
      printf("load_textures() returned %d\n", ret);
      if (ret == gsm->num_textures) {
        /* Couldn't load any textures */
        if (textures_enabled) gui_toggle_texturing();
      }
    }
    gui_generate_textures_list(gsm, 0, 0);
#if 0
    if (gsm->num_grips) {
#endif
    if (0) {
      /* weapon */
      if (weapon) cleanup_gsm(weapon);
      weapon = new gsm_t;
      memmove(weapon, gsm, sizeof(gsm_t));

      gsa_t *gsa = new gsa_t;
      create_fake_animation(*gsm, *gsa);
      if (weapon_animation) cleanup_gsa(weapon_animation);
      weapon_animation = new gsa_t;
      memmove(weapon_animation, gsa, sizeof(gsa_t));

      delete gsa;
      rotate_positions(*weapon, *weapon_animation);
    }
    else {
      /* assume character */
      if (character) cleanup_gsm(character);
      character = new gsm_t;
      memmove(character, gsm, sizeof(gsm_t));

      gsa_t *gsa = new gsa_t;
      /*
        Create a default pose if:
           There is no animation loaded
        OR This model has a different number of bones to the loaded animation
        OR The loaded animation was another model's default pose
      */
      if (! character_animation || character->num_bones != character_animation->num_bones || fake_animation) {
        create_fake_animation(*gsm, *gsa);
        frame = 0;
        if (character_animation) cleanup_gsa(character_animation);
        character_animation = new gsa_t;
        memmove(character_animation, gsa, sizeof(gsa_t));
        delete gsa;
      }
      rotate_positions(*character, *character_animation);
#if 0
      /* calculate centre of gravity */
      float highest = -9999.0f; float lowest = 9999.0f;
      unsigned short i, j;
      for (i = 0; i < character->num_positions; i++) {
        position_t &position = character->positions[i];
        if (position.world[3][1] < lowest) lowest = position.world[3][1];
        if (position.world[3][1] > highest) highest = position.world[3][1];
      }
      float diff = (highest - lowest) / 2.0f;
      for (i = 0; i < character_animation->num_bones; i++) {
        bone_t &root = character_animation->bones[i];
        for (j = 0; j < character_animation->num_frames; j++) {
          root.worlds[j][3][1] -= diff;
        }
      }
      rotate_positions(*character, *character_animation);
#endif
    }
    delete gsm;
  }
#endif
}

void gui_save_mesh(gsm_t *gsm) {
#ifdef _MSC_VER
  if (gui_fileselector("gsm", "GSkinMesh files\0*.gsm\0All files\0*.*\0\0", "Save GSKinMesh", meshdir.c_str(), 0)) {
    if (! filesel.lpstrFile) return;
    int len = strlen(filesel.lpstrFile) + 32;
    char *path = new char[len];
    _snprintf(path, len, "%s", filesel.lpstrFile);
    _snprintf(path + filesel.nFileOffset, len - filesel.nFileOffset, "%d.%d.%s", GetTickCount(), GetCurrentProcessId(), filesel.lpstrFile + filesel.nFileOffset);
    int ret = write_gsm(*gsm, path);
    if (ret) {
      sprintf(path, "Error code: %d!", ret);
      MessageBox(0, path, "Failed to write GSM", MB_OK | MB_ICONEXCLAMATION);
    }
    else {
      if (! DeleteFile(filesel.lpstrFile)) {
        ret = GetLastError();
        if (ret == ERROR_FILE_NOT_FOUND) ret = 0;
        else fprintf(stderr, "Failed to delete existing file %s: %d!\n", filesel.lpstrFile);
      }
      if (! ret && ! MoveFile(path, filesel.lpstrFile)) {
        fprintf(stderr, "Failed to rename %s to %s: %d!\n", path, filesel.lpstrFile, GetLastError());
        ret = 1;
      }
    }
    delete[] path;

    if (! ret) printf("Wrote GSM %s\n", filesel.lpstrFile);
  }
#else
  fprintf(stderr, "gui_save_mesh() unimplemented!\n");
#endif
}

void gui_remove_mesh(gsm_t *gsm) {
  if (gsm) cleanup_gsm(gsm);
#ifdef _MSC_VER
  SetDlgItemText(mesh_window, IDC_MESH_NAME, "");
  SetDlgItemText(mesh_window, IDC_NUM_POLYGONS, "");
  SetDlgItemText(mesh_window, IDC_NUM_BONES, "");
  SetDlgItemText(mesh_window, IDC_CLASS, "");
#endif
}

void gui_load_animation() {
#ifdef WIN32
  if (gui_fileselector(0, "GSkinAnimation files\0*.gsa\0Officer animations\0SWT_*.gsa\0Kid animations\0KID_*.gsa\0Male character animations\0TAC_*.gsa;BG1_*.gsa;BG2_*.gsa;BG3_*.gsa;MH1_*.gsa;MH2_*.gsa\0Female character animations\0BF1)*.gsa;FH1_*.gsa;FH2_*.gsa;PRH_*.gsa\0All files\0*.*\0\0", "Load GSKinAnimation", animdir.c_str(), &animindex)) {
    /* inefficiently allocating a gsa_t now so we can get the dirname :( */
    char *dirname = new char[filesel.nFileOffset];
    memmove(dirname, filesel.lpstrFile, filesel.nFileOffset);
    dirname[filesel.nFileOffset - 1] = '\0';
    animdir = dirname;
    /* can delete[] dirname since it isn't needed for gsa */
    delete[] dirname;

    if (! character) return;
    gsa_t *gsa = new gsa_t;

    printf("Loading mesh: %s\n", filesel.lpstrFile);
    int ret = read_gsa(*gsa, *character, filesel.lpstrFile);
    if (ret) {
      fprintf(stderr, "Error loading GSkinAnimation: %d\n", ret);
    }
    else {
      SetDlgItemText(animation_window, IDC_ANIMATION_NAME, filesel.lpstrFile + filesel.nFileOffset);
      SetDlgItemInt(animation_window, IDC_NUM_FRAMES, gsa->num_frames, 1);
      frame = 0;
      gui_set_frame_number();
      if (character_animation) cleanup_gsa(character_animation);
      character_animation = new gsa_t;
      memmove(character_animation, gsa, sizeof(gsa_t));
      delete gsa;
      
      /* This is a real animation */
      fake_animation = 0;
    }
  }
#else
  fprintf(stderr, "gui_load_animation() unimplemented!\n");
#endif
}

void gui_remove_animation(gsa_t *gsa) {
  if (gsa) cleanup_gsa(gsa);
  frame = 0;
#ifdef _MSC_VER
  SetDlgItemText(animation_window, IDC_ANIMATION_NAME, "");
  SetDlgItemText(animation_window, IDC_NUM_FRAMES, "");
  SetDlgItemText(animation_window, IDC_FRAME, "");
#endif
}

void gui_load_background() {
#ifdef WIN32
  if (gui_fileselector(0, "All files\0*.*\0\0", "Open background image", bgdir.c_str(), 0)) {
    int ret = load_background(filesel.lpstrFile);
    if (ret) fprintf(stderr, "Error loading background: %d\n", ret);
    else {
      show_background = 1;
      SetDlgItemText(menu_window, IDC_BACKGROUND_NAME, filesel.lpstrFile + filesel.nFileOffset);
    }
    filesel.lpstrFile[filesel.nFileOffset - 1] = '\0';
    bgdir = filesel.lpstrFile;
  }
#else
  fprintf(stderr, "gui_load_background() unimplemented!\n");
#endif
}

void gui_take_screenshot() {
  char *filename;
#ifdef WIN32
  if (! gui_fileselector("bmp", "Bitmap files\0*.bmp\0All files\0*.*\0\0", "Choose filename to take screenshot", screendir.c_str(), &screenindex)) return;

  filename = filesel.lpstrFile;
#endif

  int ret = take_screenshot(filename);
  if (ret) fprintf(stderr, "Error taking screenshot: %d\n", ret);
  else {
    printf("Saved screenshot at %s\n", filename);
#ifdef WIN32
    filesel.lpstrFile[filesel.nFileOffset - 1] = '\0';
    screendir = filesel.lpstrFile;
  }
#endif
}

void gui_reload_textures(gsm_t *gsm) {
  if (! gsm) return;
  if (reloading_textures) return;
  reloading_textures = 1;

  cleanup_gl_textures(gsm->gl_textures);
  int ret = load_textures(gsm->num_textures, gsm->textures, gsm->gl_textures, gsm->dirname);
  if (ret < 0) {
    fprintf(stderr, "Can't load textures!\n");
  }

  reloading_textures = 0;
}

void gui_show_textures(gsm_t *gsm) {
  if (! gsm) {
    fprintf(stderr, "No mesh loaded!\n");
    return;
  }

#ifdef WIN32
  HWND listbox = gui_get_listbox();
  if (! listbox) return;

  /* need some items */
  if (! SendMessage(listbox, LB_GETCOUNT, 0, 0)) {
    fprintf(stderr, "No textures loaded!\n");
    return;
  }

  /* of which some must be selected */
  if (! SendMessage(listbox, LB_GETSELCOUNT, 0, 0)) {
    fprintf(stderr, "No textures selected to show!\n");
    return;
  }

  int *items = new int[gsm->num_textures];
  int count = SendMessage(listbox, LB_GETSELITEMS, gsm->num_textures, (LPARAM) items);
  if (! count) {
    fprintf(stderr, "Can't get selected items: %d!\n", GetLastError());
    delete[] items;
    return;
  }

  int i;
  for (i = 0; i < gsm->num_textures; i++) {
    texture_t &texture = gsm->textures[i];
    texture.hidden = 1;
  }
  printf("Showing only selected textures:\n");
  for (i = 0; i < count; i++) {
    texture_t &texture = gsm->textures[items[i]];
    texture.hidden = 0;
    printf("  Texture %u: %s\n", items[i], texture.basename);
  }

  delete[] items;
#endif
}

void gui_toggle_shading() {
  if (culling = ! culling) glEnable(GL_CULL_FACE);
  else glDisable(GL_CULL_FACE);
  gui_update_display_options();
}

void gui_toggle_wireframe() {
  if (wireframe == GL_LINE_STRIP) wireframe = GL_TRIANGLES;
  else wireframe = GL_LINE_STRIP;
  gui_update_display_options();
}

void gui_toggle_texturing() {
  textures_enabled = ! textures_enabled;
  material_hack = ! material_hack;
  gui_update_display_options();
}

void gui_toggle_lighting() {
  if (show_hit_detect = ! show_hit_detect) glDisable(GL_LIGHTING);
  else glEnable(GL_LIGHTING);
  gui_update_display_options();
}

void gui_toggle_background() {
  show_background = ! show_background;
  gui_update_display_options();
}

void gui_toggle_normals() {
  show_normals = ! show_normals;
  gui_update_display_options();
}

void gui_toggle_mesh() {
  show_mesh = ! show_mesh;
  gui_update_display_options();
}

void gui_toggle_bones() {
  show_bones = ! show_bones;
  gui_update_display_options();
}

void gui_toggle_aa() {
  if (antialiasing = ! antialiasing) glEnable(GL_LINE_SMOOTH);
  else glDisable(GL_LINE_SMOOTH);
  gui_update_display_options();
}

#ifdef _MSC_VER
void gui_win32_paint(HWND window, HDC device, HBITMAP bmp, int item) {
  if (! bmp) return;

  RECT r;
  POINT p;

  GetClientRect(GetDlgItem(window, item), &r);
  p.x = r.left; p.y = r.top;
  ClientToScreen(GetDlgItem(window, item), &p);
  ScreenToClient(window, &p);

  DrawState(device, 0, 0, (LPARAM) bmp, 0, p.x, p.y, 0, 0, DST_BITMAP | DSS_NORMAL);
}

int CALLBACK gui_win32_mesh(HWND window, UINT message, WPARAM w, LPARAM l) {
  HDC device;
  PAINTSTRUCT ps;
  static HBITMAP load_bmp, save_bmp, remove_bmp, textures_bmp;

  switch (message) {
    case WM_INITDIALOG:
      if (mesh_window) DestroyWindow(window);
      else mesh_window = window;
      sgMakeIdentQuat(rotquat);
      /*
      load_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_LOAD_MESH));
      if (! load_bmp) fprintf(stderr, "Load GSM bitmap not loaded: %d!\n", GetLastError());
      save_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_SAVE_MESH));
      if (! save_bmp) fprintf(stderr, "Save GSM bitmap not loaded: %d!\n", GetLastError());
      remove_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_REMOVE_MESH));
      if (! remove_bmp) fprintf(stderr, "Remove GSM bitmap not loaded: %d!\n", GetLastError());
      textures_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_TEXTURES));
      if (! textures_bmp) fprintf(stderr, "Textures bitmap not loaded: %d!\n", GetLastError());
      */
      return 1;
    case WM_COMMAND:
      switch (LOWORD(w)) {
        case IDC_LOAD_MESH: gui_load_mesh(); break;
        case IDC_SAVE_MESH: gui_save_mesh(character); break;
        case IDC_REMOVE_MESH:
          gui_remove_mesh(character);
          character = 0;
          break;
        case IDC_TEXTURES: return gui_spawn_textures_window(window);
        case IDC_SHOW_BONES: gui_toggle_bones(); break;
        case IDC_SHOW_MESH: gui_toggle_mesh(); break;
        case IDC_SHOW_TEXTURES: gui_toggle_texturing(); break;
        case IDC_SHOW_LIGHTING: gui_toggle_lighting(); break;
        case IDC_SHOW_WIREFRAME: gui_toggle_wireframe(); break;
        case IDC_SHOW_AA: gui_toggle_aa(); break;
        case IDC_SHOW_BACKGROUND: gui_toggle_background(); break;
        case IDC_SHOW_NORMALS: gui_toggle_normals(); break;
      }
      return 1;
    case WM_CLOSE:
      mesh_window = 0;
      DestroyWindow(window);
      return 1;
    case WM_PAINT:
      device = BeginPaint(window, &ps);
      make_transparent(window);
      gui_win32_paint(window, device, load_bmp, IDC_LOAD_MESH);
      gui_win32_paint(window, device, save_bmp, IDC_SAVE_MESH);
      gui_win32_paint(window, device, remove_bmp, IDC_REMOVE_MESH);
      gui_win32_paint(window, device, textures_bmp, IDC_TEXTURES);
      EndPaint(window, &ps);
      break;
    case WM_WINDOWPOSCHANGED:
      /* force update of the viewer window when the menu is redrawn on top */
      draw_scene();
      break;
  }

  return 0;
}

int CALLBACK gui_win32_animation(HWND window, UINT message, WPARAM w, LPARAM l) {
  HDC device;
  PAINTSTRUCT ps;
  static HBITMAP load_bmp, remove_bmp, play_bmp, prev_bmp, next_bmp;
  gsa_t *gsa;

  switch (message) {
    case WM_INITDIALOG:
      if (animation_window) DestroyWindow(window);
      else animation_window = window;
      /*
      load_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_LOAD_ANIMATION));
      if (! load_bmp) fprintf(stderr, "Load GSA bitmap not loaded: %d!\n", GetLastError());
      remove_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_REMOVE_ANIMATION));
      if (! remove_bmp) fprintf(stderr, "Remove GSA bitmap not loaded: %d!\n", GetLastError());
      play_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_PLAY_ANIMATION));
      if (! play_bmp) fprintf(stderr, "Play animation bitmap not loaded: %d!\n", GetLastError());
      next_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_NEXT_FRAME));
      if (! next_bmp) fprintf(stderr, "Next frame bitmap not loaded: %d!\n", GetLastError());
      prev_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_PREV_FRAME));
      if (! prev_bmp) fprintf(stderr, "Previous frame bitmap not loaded: %d!\n", GetLastError());
      */
      return 1;
    case WM_COMMAND:
      switch (LOWORD(w)) {
        case IDC_LOAD_ANIMATION: gui_load_animation(); break;
        case IDC_REMOVE_ANIMATION:
          if (character) {
            gsa = new gsa_t;
            create_fake_animation(*character, *gsa);
            gui_remove_animation(character_animation);
            character_animation = gsa;
          }
          break;
        case IDC_PLAY_ANIMATION: paused = ! paused; break;
        case IDC_NEXT_FRAME: gui_next_frame(); break;
        case IDC_PREV_FRAME: gui_prev_frame(); break;
      }
      return 1;
    case WM_CLOSE:
      animation_window = 0;
      DestroyWindow(window);
      return 1;
    case WM_PAINT:
      device = BeginPaint(window, &ps);
      make_transparent(window);
      gui_win32_paint(window, device, load_bmp, IDC_LOAD_ANIMATION);
      gui_win32_paint(window, device, remove_bmp, IDC_REMOVE_ANIMATION);
      gui_win32_paint(window, device, play_bmp, IDC_PLAY_ANIMATION);
      gui_win32_paint(window, device, next_bmp, IDC_NEXT_FRAME);
      gui_win32_paint(window, device, prev_bmp, IDC_PREV_FRAME);
      EndPaint(window, &ps);
      break;
    case WM_WINDOWPOSCHANGED:
      /* force update of the viewer window when the menu is redrawn on top */
      draw_scene();
      break;
  }

  return 0;
}

int CALLBACK gui_win32_menu(HWND window, UINT message, WPARAM w, LPARAM l) {
  HWND win;
  HDC device;
  PAINTSTRUCT ps;
  static HBITMAP gsm_bmp, gsa_bmp, exit_bmp, screenshot_bmp;

  switch (message) {
    case WM_INITDIALOG: 
      gsm_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_GSM));
      if (! gsm_bmp) fprintf(stderr, "GSM bitmap not loaded: %d!\n", GetLastError());
      gsa_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_GSA));
      if (! gsa_bmp) fprintf(stderr, "GSA bitmap not loaded: %d!\n", GetLastError());
      exit_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_EXIT));
      if (! exit_bmp) fprintf(stderr, "Exit bitmap not loaded: %d!\n", GetLastError());
      screenshot_bmp = LoadBitmap(instance, MAKEINTRESOURCE(IDB_SCREENSHOT));
      if (! screenshot_bmp) fprintf(stderr, "Screenshot bitmap not loaded: %d!\n", GetLastError());
      return 1;
    case WM_COMMAND:
      switch (LOWORD(w)) {
        case IDC_MESH:
          win = CreateDialog(0, MAKEINTRESOURCE(IDD_MESH), window, gui_win32_mesh);
          if (! win) {
            if (! mesh_window) {
              fprintf(stderr, "Can't create mesh window: %d\n", GetLastError());
            }
            return 0;
          }
          gui_update_display_options();
          ShowWindow(win, SW_SHOW);
          break;
        case IDC_ANIMATION:
          win = CreateDialog(0, MAKEINTRESOURCE(IDD_ANIMATION), window, gui_win32_animation);
          if (! win) {
            if (! animation_window) {
              fprintf(stderr, "Can't create animation window: %d\n", GetLastError());
            }
            return 0;
          }
          ShowWindow(win, SW_SHOW);
          break;
        case IDC_LOAD_BACKGROUND: gui_load_background(); break;
        case IDC_SCREENSHOT: gui_take_screenshot(); break;
        case IDC_EXIT: exitasap = 1; break;
      }
      return 1;
    case WM_PAINT:
      device = BeginPaint(window, &ps);
      make_transparent(window);
      gui_win32_paint(window, device, gsm_bmp, IDC_MESH); /* XXX: why not GSM */
      gui_win32_paint(window, device, gsa_bmp, IDC_ANIMATION);
      gui_win32_paint(window, device, exit_bmp, IDC_EXIT);
      gui_win32_paint(window, device, screenshot_bmp, IDC_SCREENSHOT);
      EndPaint(window, &ps);
      break;
    case WM_WINDOWPOSCHANGED:
      /* force update of the viewer window when the menu is redrawn on top */
      draw_scene();
      break;
  }

  return 0;
}

int CALLBACK gui_win32_textures(HWND window, UINT message, WPARAM w, LPARAM l) {
  switch (message) {
    case WM_INITDIALOG:
      if (textures_window) DestroyWindow(window);
      else {
        textures_window = window;
        gui_generate_textures_list(character, 0, 0);
      }
      return 1;
    case WM_COMMAND:
      switch (LOWORD(w)) {
        case IDC_RELOAD: gui_reload_textures(character); break;
        case IDC_SELECTED: gui_show_textures(character); break;
        case IDC_BROWSE: gui_browse_for_texture(); break;
        case IDC_REPLACE: gui_replace_textures(character); break;
          return 1;
      }
      break;
    case WM_CLOSE:
      textures_window = 0;
      DestroyWindow(window);
      return 1;
    case WM_WINDOWPOSCHANGED:
      /* force update of the viewer window when the menu is redrawn on top */
      draw_scene();
      break;
  }

  return 0;
}

HWND gui_get_listbox() {
  HWND listbox = GetDlgItem(textures_window, IDC_TEXTURES);
  if (! listbox) {
    fprintf(stderr, "Can't get listbox handle: %d!\n", GetLastError());
  }
  return listbox;
}

void gui_generate_textures_list(gsm_t *gsm, int num_selected, int *selected) {
  if (! gsm) return;
  if (! textures_window) return;
  HWND listbox = gui_get_listbox();
  if (! listbox) return;

  SendMessage(listbox, LB_RESETCONTENT, 0, 0);
  bool ok = 0;
  for (unsigned short i = 0; i < gsm->num_textures; i++) {
    texture_t &texture = gsm->textures[i];
    std::string s;
    /* Is texture loaded? */
    ok = false;
    if (gsm->gl_textures.num_textures > i) {
      if (gsm->gl_textures.ok[i]) ok = true;
    }
    else {
      if (textures_enabled) {
        fprintf(stderr, "Warning: gui_generate_textures_list(): i = %u; gsm->gl_textures.num_textures = %u\n", i, gsm->gl_textures.num_textures);
      }
      else {
        /* Do nothing: textures are disabled */
      }
    }

    if (ok) s = "  ";
    else s = "! ";
    s += texture.basename;
    SendMessage(listbox, LB_ADDSTRING, 0, (LPARAM) s.c_str());
    SendMessage(listbox, LB_SETITEMDATA, i, (LPARAM) i);
  }

  if (num_selected && selected) {
    for (int j = 0; j < num_selected; j++) {
      SendMessage(listbox, LB_SELITEMRANGE, 1, MAKELPARAM(selected[j], selected[j]));
    }
  }
}

void gui_replace_textures(gsm_t *gsm) {
  if (! gsm) {
    fprintf(stderr, "No mesh loaded!\n");
    return;
  }

  HWND listbox = gui_get_listbox();
  if (! listbox) return;

  /* need some items */
  if (! SendMessage(listbox, LB_GETCOUNT, 0, 0)) {
    fprintf(stderr, "No textures loaded!\n");
    return;
  }

  /* of which some must be selected */
  if (! SendMessage(listbox, LB_GETSELCOUNT, 0, 0)) {
    fprintf(stderr, "No textures selected for replacement!\n");
    return;
  }

  /* and there must be a replacement */
  char name[MAX_PATH + 1];
  int len = GetDlgItemText(textures_window, IDC_REPLACEMENT, name, MAX_PATH);
  if (! len) {
    fprintf(stderr, "No replacement texture name!\n");
    return;
  }

  int *items = new int[gsm->num_textures];
  int count = SendMessage(listbox, LB_GETSELITEMS, gsm->num_textures, (LPARAM) items);
  if (! count) {
    fprintf(stderr, "Can't get selected items: %d!\n", GetLastError());
    delete[] items;
    return;
  }

  for (int i = 0; i < count; i++) {
    texture_t &texture = gsm->textures[items[i]];
    char *new_name = new char[len + 1];
    memmove(new_name, name, len);
    new_name[len] = '\0';
    if (texture.name_length && texture.name) delete[] texture.name;
    texture.name = new_name;
    texture.name_length = (unsigned short) len;
    set_basename(texture.name, texture.name_length, &texture.basename, &texture.basename_length);
  }

  gui_reload_textures(gsm);
  gui_generate_textures_list(gsm, count, items);
  delete[] items;
}

void gui_browse_for_texture() {
  if (gui_fileselector(0, "All supported image files\0*.bmp;*.tga\0Bitmap files\0*.bmp\0Targa files\0*.tga\0All files\0*.*\0\0", "Choose a new texture", texturedir.c_str(), &textureindex)) {
    char *dirname = new char[filesel.nFileOffset];
    memmove(dirname, filesel.lpstrFile, filesel.nFileOffset);
    dirname[filesel.nFileOffset - 1] = '\0';
    texturedir = dirname;
    delete[] dirname;

    /* set the edit box */
    if (! SetDlgItemText(textures_window, IDC_REPLACEMENT, filesel.lpstrFile)) {
      fprintf(stderr, "Can't set edit box to new texture name: %d!\n", GetLastError());
    }
  }
}

int gui_spawn_textures_window(HWND window) {
  HWND txt = CreateDialog(0, MAKEINTRESOURCE(IDD_TEXTURES), window, gui_win32_textures);
  if (! txt) {
    if (! textures_window) {
      fprintf(stderr, "Can't create textures window: %d\n", GetLastError());
    }
    return 0;
  }
  ShowWindow(txt, SW_SHOW);
  return 1;
}

#endif

void gui_update_display_options() {
#ifdef WIN32
  if (show_bones) CheckDlgButton(mesh_window, IDC_SHOW_BONES, BST_CHECKED);
  else CheckDlgButton(mesh_window, IDC_SHOW_BONES, BST_UNCHECKED);

  if (show_mesh) CheckDlgButton(mesh_window, IDC_SHOW_MESH, BST_CHECKED);
  else CheckDlgButton(mesh_window, IDC_SHOW_MESH, BST_UNCHECKED);

  if (textures_enabled) CheckDlgButton(mesh_window, IDC_SHOW_TEXTURES, BST_CHECKED);
  else CheckDlgButton(mesh_window, IDC_SHOW_TEXTURES, BST_UNCHECKED);

  if (show_hit_detect) CheckDlgButton(mesh_window, IDC_SHOW_LIGHTING, BST_CHECKED);
  else CheckDlgButton(mesh_window, IDC_SHOW_LIGHTING, BST_UNCHECKED);

  if (wireframe == GL_LINE_STRIP) CheckDlgButton(mesh_window, IDC_SHOW_WIREFRAME, BST_CHECKED);
  else CheckDlgButton(mesh_window, IDC_SHOW_WIREFRAME, BST_UNCHECKED);

  if (antialiasing) CheckDlgButton(mesh_window, IDC_SHOW_AA, BST_CHECKED);
  else CheckDlgButton(mesh_window, IDC_SHOW_AA, BST_UNCHECKED);

  if (show_background) CheckDlgButton(mesh_window, IDC_SHOW_BACKGROUND, BST_CHECKED);
  else CheckDlgButton(mesh_window, IDC_SHOW_BACKGROUND, BST_UNCHECKED);

  if (show_normals) CheckDlgButton(mesh_window, IDC_SHOW_NORMALS, BST_CHECKED);
  else CheckDlgButton(mesh_window, IDC_SHOW_NORMALS, BST_UNCHECKED);
#endif
}

void gui_set_frame_number() {
#ifdef WIN32
  SetDlgItemInt(animation_window, IDC_FRAME, frame + 1, 1);
#endif
}

void gui_next_frame() {
  if (! character_animation) return;
  if (++frame >= character_animation->num_frames) frame = 0; 
  gui_set_frame_number();
}

void gui_prev_frame() {
  if (! character_animation) return;
  if (--frame < 0) frame = character_animation->num_frames - 1;
  gui_set_frame_number();
}

void setup_gui() {
  cwd = new char[UL_NAME_MAX + 1];
  if (! ulGetCWD(cwd, UL_NAME_MAX + 1)) cwd[0] = '\0';
  int len = strlen(cwd);
  /* Default locations */
  meshdir = cwd;
  animdir = cwd;
  texturedir = cwd;
  screendir = cwd;
  /* Default filters */
  meshindex = animindex = textureindex = screenindex = 1;
#ifdef _MSC_VER
  instance = GetModuleHandle(0);
  HICON icon = LoadIcon(instance, MAKEINTRESOURCE(IDI_SWAT3));
  if (icon) SetClassLong(wm.window, GCL_HICON, (long) icon);

  /* Create menu window */
  menu_window = CreateDialog(0, MAKEINTRESOURCE(IDD_MENU), wm.window, gui_win32_menu);
  if (! menu_window) {
    fprintf(stderr, "Can't create menu window: %d\n", GetLastError());
    return;
  }

  /* Show menu window */
  ShowWindow(menu_window, SW_SHOW);

#ifdef WINDOWS2000_TRANSPARENCY_DOES_NOT_SUCK
  /* Try to load user32.dll (to get SetLayeredWindowAttributes() function) */
  library = LoadLibrary("user32.dll");
  if (! library) return;

  /* Find SetLayeredWindowAttributes() */
  function = get_function("SetLayeredWindowAttributes", library);
  printf("SetLayeredWindowAttributes(): %p\n", function);
#endif
#endif
}

void cleanup_gui() {
#ifdef WINDOWS2000_TRANSPARENCY_DOES_NOT_SUCK
  if (library) FreeLibrary(library);
#endif
}
