#include "video.h"

extern gsm_t *character, *weapon;
extern gsa_t *character_animation, *weapon_animation;
extern int width, height;
extern GLfloat xrot, yrot, zrot;
extern GLfloat xpos, ypos, zpos;
extern sgVec4 rotquat;
extern SGfloat rotangle;
extern sgVec3 rotaxis;
extern int exitasap;
extern int show_bones;
extern int show_mesh;
extern int show_background;
extern int show_normals;
extern int show_hit_detect;
extern int reverse_normals;
extern int textures_enabled;
extern int material_hack;
extern int wireframe;
extern int frame;
extern int axishack;
extern GLuint background_texture;
extern texture_image_t background;
#if 0
static Sint16 X, Y; /* Last polled mouse position */
#endif
extern SDL_SysWMinfo wm;

SDL_Surface *resize_window(int w, int h) {
  SDL_Surface *ret;
  width = w; height = h;

  printf("Creating surface: ");
  ret = SDL_SetVideoMode(width, height, 16, SDL_OPENGL | SDL_ANYFORMAT);
  if (! ret) {
    printf("failed!\n");
    return 0;
  }
  SDL_GetWMInfo(&wm);
#ifdef WIN32
  printf("%d bpp window with handle 0x%x\n", ret->format->BitsPerPixel, wm.window);
#else
  printf("%d bpp\n", ret->format->BitsPerPixel);
#endif

  /* OpenGL setup */
  if (setup_opengl(width, height)) {
    printf("OpenGL initialisation failed!\n");
    return 0;
  }
  printf("OpenGL is ready.\n");

  return ret;
}

int setup_opengl(int width, int height) {
  printf("Initialising %dx%d OpenGL window:\n", width, height);
  printf("  Shading: flat\n");
  glShadeModel(GL_FLAT);
  printf("  Lighting: ambient, local, two sided\n");
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, whitelight);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
  glLightfv(GL_LIGHT0, GL_AMBIENT, whitelight);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);
  printf("  Viewport: %dx%d\n", width, height);
  glClearColor(0.2f, 0.0f, 0.0f, 0.1f);
  glViewport(0, 0, width, height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(60.0f, (GLfloat) 4 / 3, 1.0f, 200.0f);
  printf("  Depth testing and culling etc: GL_LEQUAL depth testing, GL_CW/GL_BACK culling\n");
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glFrontFace(GL_CW);
  glCullFace(GL_BACK);
  glEnable(GL_CULL_FACE);
  printf("  Polygon mode: filled front, line back\n");
  glPolygonMode(GL_FRONT, GL_FILL);
  glPolygonMode(GL_BACK, GL_LINE);
  printf("  Point size and line width\n");
  glPointSize(4.0f);
  glLineWidth(2.0f);
  printf("  Blending\n");
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  printf("  Texturing: decal\n");
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
  return 0;
}

void draw_origin() {
  glBegin(GL_LINES);
  glColor3f(1.0f, 0.0f, 0.0f);
  glVertex3f(0.0f, 0.0f, 0.0f);
  glVertex3f(1.0f, 0.0f, 0.0f);

  glColor3f(0.0f, 1.0f, 0.0f);
  glVertex3f(0.0f, 0.0f, 0.0f);
  glVertex3f(0.0f, 1.0f, 0.0f);

  glColor3f(0.0f, 0.0f, 1.0f);
  glVertex3f(0.0f, 0.0f, 0.0f);
  glVertex3f(0.0f, 0.0f, 1.0f);
  glEnd();
}

void draw_background() {
  if (! show_background) return;
  if (exitasap) return;

  SDL_Surface *screen = SDL_GetVideoSurface();
  if (! screen) {
    fprintf(stderr, "Can't get screen to draw background!\n");
    return;
  }

  glPushAttrib(GL_ENABLE_BIT | GL_VIEWPORT_BIT | GL_TRANSFORM_BIT);
  glDisable(GL_LIGHTING);
  glDisable(GL_DEPTH_TEST);

  glViewport(0, 0, screen->w, screen->h);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluOrtho2D(0, screen->w, 0, screen->h);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  glColor4fv(whitelight);
  glBindTexture(GL_TEXTURE_2D, background_texture);
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_CULL_FACE);
  glColor4fv(whitelight);
  glBegin(GL_QUADS);
  glTexCoord2d(0.0, 0.0);
  glVertex2i(0, 0);

  glTexCoord2d(0.0, 1.0);
  glVertex2i(0, screen->h);

  glTexCoord2d(1.0, 1.0);
  glVertex2i(screen->w, screen->h);

  glTexCoord2d(1.0, 0.0);
  glVertex2i(screen->w, 0);
  glEnd();

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glPopAttrib();

  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
}

void draw_scene() {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  draw_background();

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(xpos, ypos, zpos);
  sgQuatToAngleAxis(&rotangle, rotaxis, rotquat);
  glRotatef(rotangle, rotaxis[0], rotaxis[1], rotaxis[2]);

  //draw_origin();

  glColor4fv(whitelight);
  if (character) draw_gsm(*character, *character_animation);
  if (weapon) draw_gsm(*weapon, *weapon_animation);

  SDL_GL_SwapBuffers();
}

void poll_joysticks() {
  SDL_Joystick *joystick;
  GLfloat value;
  for (int i = 0; i < SDL_NumJoysticks(); i++) {
    joystick = SDL_JoystickOpen(i);
    if (! joystick) continue;

    value = (GLfloat) SDL_JoystickGetAxis(joystick, 0);
    /* horiz */
    if (fabs(value) > DEADZONE) {
      sgRotQuat(rotquat, value / JOYSTICKSPEED, 0.0f, 0.0f, 1.0f);
    }
    /* vert */
    value = (GLfloat) SDL_JoystickGetAxis(joystick, 1);
    if (fabs(value) > DEADZONE) {
      sgRotQuat(rotquat, value / JOYSTICKSPEED, 1.0f, 0.0f, 0.0f);
    }
    /* rudder */
    value= (GLfloat) SDL_JoystickGetAxis(joystick, 3);
    if (fabs(value) > DEADZONE) {
      sgRotQuat(rotquat, value / JOYSTICKSPEED, 0.0f, 1.0f, 0.0f);
    }

    SDL_JoystickClose(joystick);
  }
}

void poll_mouse(Uint8 buttons, Sint16 x, Sint16 y) {
  if (! buttons) return;

#ifdef SDL_BUTTON_WHEELUP
  if (buttons & SDL_BUTTON(SDL_BUTTON_WHEELUP)) {
    zpos += MOUSEPOSSPEED;
  }
  else if (buttons & SDL_BUTTON(SDL_BUTTON_WHEELDOWN)) {
    zpos -= MOUSEPOSSPEED;
  }
#endif

  if (buttons & SDL_BUTTON(SDL_BUTTON_MIDDLE) || axishack) {
    /* Middle button down so rotate about Z or zoom */
    if (buttons & SDL_BUTTON(SDL_BUTTON_LEFT)) {
      sgRotQuat(rotquat, (GLfloat) x / ((GLfloat) width / MOUSEROTSPEED), 0.0f, 0.0f, 1.0f);
    }
    if (buttons & SDL_BUTTON(SDL_BUTTON_RIGHT)) {
      zpos += (GLfloat) y / ((GLfloat) height / MOUSEPOSSPEED);
      zpos += (GLfloat) x / ((GLfloat) width / MOUSEPOSSPEED);
    }
  }
  else {
    if (buttons & SDL_BUTTON(SDL_BUTTON_LEFT)) {
      /* Rotate about X or Y */
      sgRotQuat(rotquat, (GLfloat) x / ((GLfloat) width / MOUSEROTSPEED), 0.0f, 1.0f, 0.0f);
      sgRotQuat(rotquat, (GLfloat) y / ((GLfloat) height / MOUSEROTSPEED), 1.0f, 0.0f, 0.0f);
    }
    if (buttons & SDL_BUTTON(SDL_BUTTON_RIGHT)) {
      /* Pan */
      xpos += (GLfloat) x / ((GLfloat) width / MOUSEPOSSPEED);
      ypos -= (GLfloat) y / ((GLfloat) height / MOUSEPOSSPEED);
    }
  }
}

#if 0
void handle_button_press(Uint8 button, Uint16 x, Uint16 y) {
  X = (Uint8) x; Y = (Uint8) y;
}

void handle_button_release(Uint8 button, Uint16 x, Uint16 y) {
}
#endif

void draw_gsm(gsm_t &gsm, gsa_t &gsa) {
  unsigned short i, j;
  int k;

  //printf("drawing %s at frame %u\n", gsm.name, frame);
  if (frame >= gsa.num_frames) {
    printf("PANIC! trying to draw %s frame %u but there are only %u!\n", gsm.name, frame, gsa.num_frames);
    frame = 0;
  }
  
  /* draw mesh */
  if (show_mesh) {
    rotate_positions(gsm, gsa);

    for (i = 0; i < gsm.num_groups; i++) {
      group_t &group = gsm.groups[i];
      material_t &material = gsm.materials[group.material_id];
      if (magic_material_name(&material)) continue;

      int texture_ok = 0;
      GLuint texture_id;
      if (material_hack) {
        sgVec4 l;
        sgCopyVec4(l, whitelight);
        l[3] = material.gl_opacity;
        glMaterialfv(material.face, GL_AMBIENT, l);
        glMaterialfv(material.face, GL_DIFFUSE, l);
        glMaterialfv(material.face, GL_SPECULAR, l);
        glMaterialfv(material.face, GL_EMISSION, l);
        glMaterialf(material.face, GL_SHININESS, 1.0f);
      }
      else {
        glMaterialfv(material.face, GL_AMBIENT, material.gl_ambient);
        glMaterialfv(material.face, GL_DIFFUSE, material.gl_diffuse);
        glMaterialfv(material.face, GL_SPECULAR, material.gl_specularity);
        glMaterialfv(material.face, GL_EMISSION, material.gl_emission);
        glMaterialf(material.face, GL_SHININESS, material.gl_shininess);
      }
      if (material.num_maps) {
        map_t &map = material.maps[0];
        texture_id = (GLuint) map.texture_id;
        if (gsm.textures[texture_id].hidden) continue;
        if (texture_id < gsm.gl_textures.num_textures) {
          if (gsm.gl_textures.ok[texture_id]) {
            texture_ok = 1 & textures_enabled;
            glBindTexture(GL_TEXTURE_2D, gsm.gl_textures.textures[texture_id]);
            glEnable(GL_TEXTURE_2D);
          }
        }
      }
      if (! texture_ok) glDisable(GL_TEXTURE_2D);

      for (j = 0; j < group.num_triangles; j++) {
        triangle_t &triangle = gsm.triangles[group.first_triangle + j];
        sgVec3 vertices[3];

        /* first calculate the points */
        for (k = 0; k < 3; k++) {
          vertex_t &vertex = gsm.vertices[triangle[k]];
          if (vertex.position_id >= gsm.num_positions) {
            find_blend(vertices[k], vertex.position_id - gsm.num_positions, gsm.blended, gsm.positions);
          }
          else {
            position_t &position = gsm.positions[vertex.position_id];
            sgCopyVec3(vertices[k], position.world[3]);
          }
        }

        /* then calculate the normal */
        sgVec3 normal;
        if (reverse_normals) sgMakeNormal(normal, vertices[0], vertices[1], vertices[2]);
        else sgMakeNormal(normal, vertices[2], vertices[1], vertices[0]);
        glNormal3fv(normal);

        /* draw it */
        if (show_normals) {
          glColor4f(1.0f, 1.0f, 0.0f, 0.0f);
          glDisable(GL_TEXTURE_2D);
          glDisable(GL_LIGHTING);
          glBegin(GL_LINES);
          sgVec3 centre;
          find_triangle_centre(centre, vertices[0], vertices[1], vertices[2]);
          glVertex3fv(centre);
          glVertex3f(centre[0] + normal[0], centre[1] + normal[1], centre[2] + normal[2]);
          glEnd();
          if (textures_enabled) glEnable(GL_TEXTURE_2D);
          if (! show_hit_detect) glEnable(GL_LIGHTING);
          glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
        }

        /* do triangles and UVs */
        glBegin(wireframe);
        for (k = 0; k < 3; k++) {
          vertex_t &vertex = gsm.vertices[triangle[k]];
          if (texture_ok) {
            uv_t &uv = gsm.uvs[vertex.uv_id];
            glTexCoord2f(uv[0], -uv[1]);
          }
          if (show_hit_detect) hit_detect_colour(vertex.hit_detect);
          glVertex3fv(vertices[k]);
        }
        glEnd();
      }
    }
  }

  /* draw skeleton */
  if (show_bones) {
    glClear(GL_DEPTH_BUFFER_BIT);
    sgVec4 l = { 1.0f, 1.0f, 0.0f, 0.5f };
    glColor4fv(whitelight);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, l);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, l);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, l);
    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, l);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 0.5f);
    /* joints */
    glBegin(GL_POINTS);
    for (i = 0; i < gsm.num_bones; i++) {
      bone_t &bone = gsa.bones[i];
      glVertex3fv(bone.worlds[frame][3]);
    }
    glEnd();
    /* bones */
    glBegin(GL_LINES);
    for (i = 0; i < gsm.num_bones; i++) {
      bone_t &bone = gsa.bones[i];
      if (bone.parent == 0xffff) continue;
      bone_t &parent = gsa.bones[bone.parent];
      glVertex3fv(parent.worlds[frame][3]);
      glVertex3fv(bone.worlds[frame][3]);
    }
    glEnd();
    /* grippers */
    draw_grippers(gsm, gsa);
  }
}

void rotate_positions(gsm_t &gsm, gsa_t &gsa) {
  static bool once = false;
  unsigned short f = frame;
  if (&gsm == weapon) f = 0;
  /* calculate world co-ordinates for positions */
  for (unsigned short i = 0; i < gsm.num_positions; i++) {
    position_t &position = gsm.positions[i];
    bone_t &bone = gsa.bones[position.bone_id];
    sgMultMat4(position.world, bone.worlds[f], position.local);
#if 0
    if (! once) {
      /* See if we can construct the world -> local matrix */
      sgMat4 m;
      sgTransposeNegateMat4(m, bone.worlds[f]);
      sgMat4 n;
      sgMultMat4(n, m, position.world);
      printf("       [ % 0.2f, % 0.2f, % 0.2f, % 0.2f ]         [ % 0.2f, % 0.2f, % 0.2f, % 0.2f ]\n", position.local[0][0], position.local[1][0], position.local[2][0], position.local[3][0], n[0][0], n[1][0], n[2][0], n[3][0]);
      printf("Actual [ % 0.2f, % 0.2f, % 0.2f, % 0.2f ]  Calc'd [ % 0.2f, % 0.2f, % 0.2f, % 0.2f ]\n", position.local[0][1], position.local[1][1], position.local[2][1], position.local[3][1], n[0][1], n[1][1], n[2][1], n[3][1]);
      printf(" local [ % 0.2f, % 0.2f, % 0.2f, % 0.2f ]   local [ % 0.2f, % 0.2f, % 0.2f, % 0.2f ]\n", position.local[0][2], position.local[1][2], position.local[2][2], position.local[3][2], n[0][1], n[1][1], n[2][2], n[3][2]);
      printf("       [ % 0.2f, % 0.2f, % 0.2f, % 0.2f ]         [ % 0.2f, % 0.2f, % 0.2f, % 0.2f ]\n", position.local[0][3], position.local[1][3], position.local[2][3], position.local[3][3], n[0][3], n[1][3], n[2][3], n[3][3]);
      once = true;
    }
#endif
  }
}

void rotate_grippers(unsigned short num_grippers, gripper_t *grippers, gsa_t &gsa, bool is_grip) {
  for (unsigned short i = 0; i < num_grippers; i++) {
    gripper_t &gripper = grippers[i];
    bone_t &bone = gsa.bones[gripper.bone_id];
    sgMat4 m;
    if (is_grip) {
      /* first fix the co-ordinate system */
      for (unsigned int j = 0; j < 3; j++) {
        for (unsigned int k = 0; k < 3; k++) {
          m[j][k] = gripper.local[k][j];
        }
        m[3][j] = -gripper.local[3][j];
      }
      /* translation is in local co-ordinates */
      sgXformVec3(m[3], m[3], m);
    }
    else memmove(m, gripper.local, sizeof(sgMat4));
    sgMultMat4(gripper.world, bone.worlds[frame], m);
  }
}

#define vertex(x, y, z) g[0] = (float) x; g[1] = (float) y; g[2] = (float) z; sgXformPnt3(w[i++], g, gripper.world)
#define triangle(x, y, z) glVertex3fv(w[x]); glVertex3fv(w[y]);   glVertex3fv(w[z])
static void draw_grip_shape(gripper_t &gripper) {
  sgVec3 g, w[19];
  unsigned int i = 0;

  vertex(-1.1359, -2.8632, -0.0656);
  vertex(0.4591, -3.6381, -0.5676);
  vertex(0.4591, -3.6381, 0.4365);
  vertex(1.4661, -1.5421, -0.5676);
  vertex(1.4661, -1.5421, 0.4365);
  vertex(3.7403, -1.5400, -0.3736);
  vertex(3.7403, -1.5400, 0.2425);
  vertex(4.4537, -0.0745, -0.5676);
  vertex(4.4537, -0.0745, 0.4365);
  vertex(0.3762, -0.3598, -0.0656);
  vertex(0.3849, 2.5184, -0.0656);
  vertex(5.5816, 2.5396, 0.4365);
  vertex(5.5816, 2.5396, -0.5676);
  vertex(6.6919, 1.0046, 0.4365);
  vertex(6.6919, 1.0046, -0.5676);
  vertex(10.6446, 0.9875, 0.4365);
  vertex(10.6446, 0.9875, -0.5676);
  vertex(10.6534, 2.5208, -0.5676);
  vertex(10.6534, 2.5208, 0.4365);

  glBegin(GL_TRIANGLES);
  triangle(2, 1, 0);
  triangle(4, 3, 1);
  triangle(2, 4, 1);
  triangle(6, 5, 3);
  triangle(4, 6, 3);
  triangle(1, 3, 0);
  triangle(0, 3, 9);
  triangle(4, 2, 0);
  triangle(4, 0, 9);
  triangle(4, 9, 8);
  triangle(4, 8, 6);
  triangle(9, 3, 7);
  triangle(3, 5, 7);
  triangle(6, 8, 7);
  triangle(7, 5, 6);
  triangle(11, 10, 12);
  triangle(11, 8, 9);
  triangle(9, 10, 11);
  triangle(10, 9, 7);
  triangle(7, 12, 10);
  triangle(17, 16, 15);
  triangle(15, 18, 17);
  triangle(14, 7, 8);
  triangle(8, 13, 14);
  triangle(13, 8, 11);
  triangle(7, 14, 12);
  triangle(16, 14, 13);
  triangle(13, 15, 16);
  triangle(17, 12, 14);
  triangle(14, 16, 17);
  triangle(18, 11, 12);
  triangle(12, 17, 18);
  triangle(15, 13, 11);
  triangle(11, 18, 15);
  glEnd();
}

void draw_throwable_shape(gripper_t &gripper) {
  sgVec3 g, w[8];
  unsigned int i = 0;

  vertex(-0.773873, -0.773873, 0.0);
  vertex(0.773873, -0.773873, 0.0);
  vertex(0.773873, 0.773873, 0.0);
  vertex(-0.773873, 0.773873, 0.0);
  vertex(-0.773873, -0.773873, 5.904398);
  vertex(0.773873, -0.773873, 5.904398);
  vertex(0.773873, 0.773873, 5.904398);
  vertex(-0.773873, 0.773873, 5.904398);

  glBegin(GL_TRIANGLES);
  triangle(0, 1, 2);
  triangle(0, 2, 3);
  triangle(2, 1, 6);
  triangle(1, 5, 6);
  triangle(2, 6, 7);
  triangle(2, 7, 3);
  triangle(0, 3, 4);
  triangle(3, 7, 4);
  triangle(1, 0, 5);
  triangle(0, 4, 5);
  triangle(4, 5, 7);
  triangle(6, 7, 5);
  glEnd();
}
#undef vertex
#undef triangle

void draw_gripper(gripper_t &gripper) {
  /* (very) crude check for Grip2 */
  if (gripper.name[gripper.name_length - 1] == '2') draw_throwable_shape(gripper);
  else draw_grip_shape(gripper);
}

void draw_grippers(gsm_t &gsm, gsa_t &gsa) {
  /* set gripper world co-ordinates */
  rotate_grippers(gsm.num_grippers, gsm.grippers, gsa, false);
  rotate_grippers(gsm.num_grips, gsm.grips, gsa, true);

  unsigned short i;
  for (i = 0; i < gsm.num_grippers; i++) {
    draw_gripper(gsm.grippers[i]);
    glBegin(GL_LINES);
    glVertex3fv(gsa.bones[gsm.grippers[i].bone_id].worlds[frame][3]);
    glVertex3fv(gsm.grippers[i].world[3]);
    glEnd();
  }
  for (i = 0; i < gsm.num_grips; i++) {
    draw_gripper(gsm.grips[i]);
    glBegin(GL_LINES);
    glVertex3fv(gsa.bones[gsm.grips[i].bone_id].worlds[frame][3]);
    glVertex3fv(gsm.grips[i].world[3]);
    glEnd();
  }
}

void hit_detect_colour(unsigned short rgb) {
  glColor4f((((rgb >> 10) % 32) * 8) / 255.0f, (((rgb >> 5) % 32) * 8) / 255.0f, ((rgb % 32) * 8) / 255.0f, 1.0f);
}
